/* DChub - a Direct Connect text clone for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * hub_cluster.h: Copyright (C) Uriel Chemouni <uriel@xor.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <glib.h>

#include "gvar.h"
#include "cnx_lst.h"
#include "manage_cnx.h"
#include "timed_out_string.h"
#include "tos_key.h"
#include "hub_cluster.h"
#include "emb_perl.h"
#include "network.h"
#include "main.h"

void free_cmd_glist(GList **pa);
void free_cmd_glist_ref(GList **pa);

int connecttome_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param);
int revconnecttome_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param);
int version_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param);
int myinfo_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param);
int private_chat_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param);
static const char *lst_cnx_type[]=
										{
											"56Kbps",
											"33.6Kbps",
											"28.8Kbps",
											"Satellite",      /* according to DC, passive mode is required here */
											"ISDN",
											"DSL",
											"Cable",
											"LAN(T1)",
											"LAN(T3)",
											NULL
											};

static const char *fnd_static_cnx_type(char *type)
{
	int i;

	i=0;
	while(lst_cnx_type[i]!=NULL)
	{
		if(!strcmp(lst_cnx_type[i],type))
			return lst_cnx_type[i];
		i++;
	}

	return "";
}

t_route_table *route_table = 0;

t_route_table *new_route_table(void)
{
	t_route_table *table;

	table = malloc(sizeof(t_route_table));

	table->server_count = 1;
	table->server_allocated_count = 10;
	table->max_server_id = 0;
	table->table = malloc((1 + table->server_allocated_count) * sizeof(CNX_ENTRY)); /* securit : + 1 */
	bzero(table->table, (1 + table->server_allocated_count) * sizeof(CNX_ENTRY));
	return (table);
}
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>

#include "color_debug.h"

int	parse_new_entry_from_hub(GString *inp, CNX_ENTRY *ptr)
{
	char	*str;
	char	*t,*t2;
	CNX_ENTRY *new_ent;
	
	new_ent = malloc(sizeof (CNX_ENTRY));
	bzero(new_ent, sizeof (CNX_ENTRY));
	t = str = strdup(inp->str + strlen("new:"));

	goto start_parsse;
	
 have_perr:
	printf("have_perr:\n");
	if(new_ent->user_nick)
		g_string_free(new_ent->user_nick, TRUE);
	if(new_ent->user_description)
		g_string_free(new_ent->user_description, TRUE);
	if(new_ent->user_mail)
		g_string_free(new_ent->user_mail, TRUE);
	if(new_ent->client_version)
		g_string_free(new_ent->client_version, TRUE);
	free(str);
	free(new_ent);
	g_string_free(inp,TRUE);
	return(0);
	
 start_parsse:
	new_ent->sock_fd=0;
	new_ent->start_time=time(NULL);
	new_ent->handshake_done=1; 
	new_ent->user_flag=1;
	new_ent->server_id = ptr->server_id;

	/* extract the nickname */
	t2 = t;
	if(*t2=='\0')
		goto have_perr;
	t2=strchr(t,'$'); /* OKI nick		1 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;

	new_ent->user_nick=g_string_new(t);
	t=t2+1;
	
	t2=strchr(t,'$'); /* OKI dexc		2 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;
	new_ent->user_description=g_string_new(t);
	t=t2+1;
	
	t2=strchr(t,'$'); /* oki mail		3 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;
	new_ent->user_mail=g_string_new(t);
	t=t2+1;
	
	t2=strchr(t,'$'); /* oki cnx_type		4 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;
	new_ent->user_cnx_type=fnd_static_cnx_type(t);
	t=t2+1;
	
	t2=strchr(t,'$');/* oki user flag			5 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;
	new_ent->user_flag=*t;
	t=t2+1;
	
	t2=strchr(t,'$'); /* OKI share size				 	6 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;
	sscanf(t,"%Lu",&(new_ent->shared_size));
	t=t2+1;
	
	t2=strchr(t,'$'); /* oki share IP		7 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;
	new_ent->ip.s_addr=inet_network(t);
	t=t2+1;
	
	t2=strchr(t,'$'); /* oki share is_op		8 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;
	new_ent->is_op=atoi(t);
	t=t2+1;
	
	t2=strchr(t,'$'); /* OKI client_version		9 */
	if(t2==NULL)
		goto have_perr;
	*t2 = 0;
	new_ent->client_version=g_string_new(t);
	t=t2+1;
	new_ent->sock_fd = -1;
	free(str);
	if (get_in_cnx_hash_table_by_nick(cnx_lst, new_ent->user_nick->str))
	{
		if (inp->str[2] == '2') /* parsing a ne2 cmd the override the user with the same user name*/
		{
			printf(COLOYEL1"doublon effacer %s "COLONONE"\n", inp->str);
			del_in_cnx_hash_table_by_nick(cnx_lst, new_ent->user_nick->str); /*remove the older one*/
			add_in_cnx_hash_table(cnx_lst, new_ent);
		}
		else
			delete_cnx_unlinked_entry(new_ent); /* delete the new one */
	}
	else
		add_in_cnx_hash_table(cnx_lst, new_ent);
	g_string_free(inp,TRUE);
	return (0);
}

int parse_pvm_from_hub(GString *inp, CNX_ENTRY *ptr)
{
	char	*user_name;
	char	*pos;
	int	ret=0;
	CNX_ENTRY	*virtual_ent;
 
	inp = g_string_erase(inp, 0, strlen("pvm:"));
	pos = strchr(inp->str, '$');
	if (pos == NULL)
		return(0);
	user_name = alloca(pos - inp->str+ 1);
	memcpy(user_name, inp->str, pos - inp->str);
	user_name[pos - inp->str] = 0;
	
	virtual_ent=get_in_cnx_hash_table_by_nick(cnx_lst, user_name);
	if (virtual_ent == NULL)
		return(0);
	inp = g_string_erase(inp, 0, strlen(user_name) + 1);
	if (!strncmp(inp->str, "$RevConnectToMe ", strlen("$RevConnectToMe ")))
		ret = revconnecttome_process("$RevConnectToMe ", virtual_ent, inp, NULL);
	else if (!strncmp(inp->str, "$To: ", strlen("$To: ")))
	{
		ret = private_chat_process("$To ", virtual_ent, inp, NULL);
	}
	else if (!strncmp(inp->str, "$ConnectToMe ", strlen("$ConnectToMe ")))
	{
		/* printf ("interconnect $ConnectToMe \n"); */
		ret = connecttome_process("$ConnectToMe ", virtual_ent, inp, NULL);
	}
	else if (!strncmp(inp->str, "$RevConnectToMe ", strlen("$RevConnectToMe ")))
	{
		/* printf ("interconnect $RevConnectToMe\n"); */
		ret = revconnecttome_process("$RevConnectToMe ", virtual_ent, inp, NULL);
	}
	else
	{
		printf (COLOMAJ1"dont know how to treater %s"COLONONE"\n", inp->str);
		g_string_free(inp,TRUE);
	}
	return(ret);
}

int parse_rem_entry_from_hub(GString *inp, CNX_ENTRY *ptr)
{
	return (0);
}

int parse_brd_from_hub(GString *inp, CNX_ENTRY *ptr)
{ 
	int ret;

	inp = g_string_erase(inp, 0, strlen("brd:"));
	if (!strncmp(inp->str, "$Version ", strlen("$Version ")))
		ret = version_process("$Version ", ptr, inp, NULL);
	else if (!strncmp(inp->str, "$MyINFO ", strlen("$MyINFO ")))
		ret = myinfo_process("$MyINFO ", ptr, inp, NULL);
	else if (!strncmp(inp->str, "$Quit ", strlen("$Quit ")))
	{
		GString_ref *local_str;
			
		local_str=g_string_ref_new(inp->str);
		queue_to_local_everyone(local_str);
		g_string_ref_free(local_str);
			
		inp = g_string_truncate (inp, inp->len - 1);
		del_in_cnx_hash_table_by_nick(cnx_lst, inp->str + strlen("$Quit "));
		ret = 0;
		g_string_free(inp,TRUE);
	}
	else
	{
		int i;
		CNX_ENTRY *ent;
		GString_ref	*inp2;
			
		inp2 = g_string_ref_new(inp->str);
		for(i=0;i<HASH_SIZE;i++)
		{
			for(ent=cnx_lst->entry[i];ent;ent=ent->next)
			{
				if((ent->sock_fd!=-1)&&(ent->server_id==0))
					send_str_to(ent,inp2);
			}
		}
			
		for(ent=cnx_lst->wait; ent; ent=ent->next) 
		{
			if((ent->sock_fd!=-1)&&(ent->user_nick!=NULL)&&(ent->user_nick->len!=0)) // MINIFIXME
				send_str_to(ent,inp2);
		}
		ret = 0;
		g_string_ref_free(inp2);
		g_string_free(inp,TRUE);
	}
	return (ret);
}

int allocate_new_hub(CNX_ENTRY *ptr)
{
	int	i;

	for (i = 1; i <= route_table->max_server_id; i++)
	{
		if (route_table->table[i].connected == 0)
		{
			if (route_table->table[i].user_flag == 0)
			{
				route_table->table[i].user_flag = 1;
				ptr->server_id = i;
			}
			else
			{
				printf(COLOBLEU1"serveur %d reserved"COLONONE"\n", i);
			}
		}
		else
		{
			printf(COLOBLEU1"serveur %d in use"COLONONE"\n", i);	
		}
	}

	printf(COLOBLEU1"%d will be used"COLONONE"\n", i);
	if (i > route_table->max_server_id)
	{
		if (route_table->max_server_id >= route_table->server_allocated_count)
		{
			printf("try to realloc HUB\n");
			route_table->server_allocated_count += 5;
			route_table->table = realloc(route_table->table, (route_table->server_allocated_count + 1) * sizeof(CNX_ENTRY));
		}
		ptr->server_id = i;
		route_table->max_server_id = i;
	}
	return (i);
}

int	interhub_cnx(char *hub_ip, int type)
{
	CNX_ENTRY *ptr;
	char	*port_str;
	int	port;
	int	socket_fd;

	printf(COLOMAJ1"interhub_cnx RCV %s"COLONONE"\n", hub_ip);

	port_str = strchr(hub_ip,':');
	if (port_str == NULL)
		port = 411;
	else
	{
		port=atoi(port_str+1);
		if((port<1)||(port>65535))
		{
			printf("invalid port recived form other hub\n");
			return(-1);
		}
		*port_str = 0;
	}

	socket_fd=create_and_open_sock_on(hub_ip,port,0,gl_bind_addr);
	if (socket_fd < 0)
	{
		printf("connection to %s:%hu fail\n", hub_ip, port);
		return(-1);
	}
	ptr=malloc(sizeof(CNX_ENTRY));
	if(ptr==NULL)
	{
		printf("Out of memory, malloc fail.\n");
		return(-3);
	}
	printf(COLOMAJ1"init strucure for new hub"COLONONE"\n");
	bzero(ptr, sizeof(CNX_ENTRY));
	ptr->sock_fd=socket_fd;
	ptr->start_time=time(NULL);
	ptr->is_op=HUB_PRIV; /* hub type */
	ptr->user_flag = type; /* CNX TYPE II */
	ptr->log_level = -5;
	ptr->server_id = allocate_new_hub(ptr);
	printf("fin interhub_cnx\n");
	add_in_cnx_wait_table(cnx_lst, ptr);
	return(0);
}

int	parse_hub_from_hub(GString *inp, CNX_ENTRY *ent)
{
	printf(COLOMAJ1"RCV HUB fro HUB (%s)"COLONONE"\n", inp->str);
	interhub_cnx(inp->str + sizeof("hub:")-1, 2);
	g_string_free(inp,TRUE);
	return (0);
}

/*
entraine un ba de l'autre cot:
 add_tos_entry_v1_uniq(TBAN_TOSKEY,duration,(void*)(&(ptr->ip)),sizeof(struct in_addr),NULL,0);
*/

int send_ban_entry_to_hub(CNX_ENTRY* ent, unsigned int duration)
{
	GString	*paked_str;
	GString_ref	*paked_str2;
	int		i;

	paked_str = g_string_new("");
	g_string_sprintf(paked_str,"ban:%s$%u|",ent->user_nick->str, duration);
	paked_str2=conv_gstring_to_gstring_ref(paked_str);

	/* printf("> sending ban entry %s\n", paked_str->str); */
	for (i = 1; i <= route_table->max_server_id; i++)
	{
		if (route_table->table[i].connected != 0)
		{
			/* printf("> sending sur %d %s\n",i , paked_str->str); */
			send_str_to((route_table->table + i), paked_str2);
		}
	}
	g_string_ref_free(paked_str2);
	return (0);
}

int	parse_ban_from_hub(GString *inp, CNX_ENTRY *ent)/* "ban:%s$%u|" */
{
	char	*str;
	char	*t;
	unsigned duration;
	CNX_ENTRY *ptr;

	str = inp->str + sizeof("ban:")-1;
	t=strchr(str,'$');
	if (t != NULL)
	{
		*t = '\0';
		t++;
		duration = atoi(t);
		ptr=find_cnx_by_nickname(str);
		if (ptr)
		{
			add_tos_entry_v1_uniq(TBAN_TOSKEY,duration,(void*)(&(ptr->ip)),sizeof(struct in_addr),NULL,0);
		}
		printf("User %s banned for %d minutes\n", ptr->user_nick->str, duration/60);
	}
	g_string_free(inp,TRUE);
	return (0);
}


/*
entraine un de l'autre cot
cnx_io_processing(ent,1,1);
shutdown(ent->sock_fd,0);	
*/

int send_kick_entry_to_hub(CNX_ENTRY* ent)
{
	GString	*paked_str;
	GString_ref	*paked_str2;
	CNX_ENTRY*	dest;

	paked_str = g_string_new("");
	g_string_sprintf(paked_str,"kik:%s|",ent->user_nick->str);
	paked_str2=conv_gstring_to_gstring_ref(paked_str);
	dest = route_table->table + ent->server_id;
	if (dest->connected == 1)
		send_str_to(dest, paked_str2);
	g_string_ref_free(paked_str2);
	return(0);
}

int parse_kik_from_hub(GString *inp, CNX_ENTRY *ent) /* "kik:%s|" */
{
	char	*str;
	CNX_ENTRY *ptr;

	str=strchr(inp->str,'|');
	if (str)
	{
		*str = 0;
		ptr=find_cnx_by_nickname(inp->str + sizeof("kik:")-1);
		if (ptr && ptr->sock_fd != -1)
		{
			cnx_io_processing(ptr,1,1);
			shutdown(ptr->sock_fd,0);
		}
	}
	g_string_free(inp,TRUE);
	return (0);
}


int	send_new_entry_to_hub(CNX_ENTRY* new_ent, CNX_ENTRY* dest, int overwrite)
{
	GString	*paked_str;
	GString_ref	*paked_str2;
	const char	*format;
#ifdef __CYGWIN__
	char		long_int[48];
#endif
	
#ifdef __CYGWIN__
	format = (overwrite==1)?"ne2:%s$%s$%s$%s$%c$%s$%s$%d$%s$|":"new:%s$%s$%s$%s$%c$%s$%s$%d$%s$|";
	sprintf(long_int, "%Lu", new_ent->shared_size);
#else
	format = (overwrite==1)?"ne2:%s$%s$%s$%s$%c$%Lu$%s$%d$%s$|":"new:%s$%s$%s$%s$%c$%Lu$%s$%d$%s$|";
#endif

	paked_str = g_string_new("");
	g_string_sprintf(paked_str,format,
			 new_ent->user_nick->str,		/*1 s	*/
			 new_ent->user_description->str,	/*2 s	*/
			 new_ent->user_mail->str,		/*3 s	*/
			 new_ent->user_cnx_type,		/*4 s	*/
			 (char)new_ent->user_flag,		/*5 c	*/
#ifdef __CYGWIN__
			 long_int,		/*6 Lu */
#else
			 new_ent->shared_size,		/*6 Lu */
#endif
			 inet_ntoa(new_ent->ip),		/*7 s	*/
			 new_ent->is_op,			/*8 d	*/
			 new_ent->client_version->str);	/*9 s */
	paked_str2=conv_gstring_to_gstring_ref(paked_str);
	send_str_to(dest, paked_str2);
	g_string_ref_free(paked_str2);
	return(0);
}

int lost_hub(CNX_ENTRY* ptr)
{
	GString *msg;
	GString_ref *msg2;
	int	i;

	printf(COLOYEL1"debut lost hub"COLONONE"\n");
	for(i=0;i<HASH_SIZE;i++) /* for getnicklist_process */
	{
		CNX_ENTRY *ent;

		struct_altered:
		for(ent=cnx_lst->entry[i]; ent; ent=ent->next)
		{
			if (ent->server_id != ptr->server_id)
				continue;
			msg=g_string_new("");
			g_string_sprintf(msg,"$Quit %s|",ent->user_nick->str);
			msg2 = conv_gstring_to_gstring_ref(msg);
			queue_to_everyone(msg2);
			send_evt_to_script("quit",ent->user_nick->str,0);
			g_string_ref_free(msg2);
			printf("cnx over4.\n");
			delete_cnx_entry(ent);
			goto struct_altered;
		}
	}

	if(ptr->sock_fd!=-1)
	{
		close(ptr->sock_fd);
		ptr->sock_fd=-1;
	}

	if(ptr->user_nick!=NULL)
		g_string_free(ptr->user_nick,TRUE);
	if(ptr->user_mail!=NULL)
		g_string_free(ptr->user_mail,TRUE);
	if(ptr->user_description!=NULL)
		g_string_free(ptr->user_description,TRUE);
	if(ptr->client_version!=NULL)
		g_string_free(ptr->client_version,TRUE);
	if(ptr->incoming_commands!=NULL)
		free_cmd_glist(&(ptr->incoming_commands));
	if(ptr->outgoing_commands!=NULL)
		free_cmd_glist_ref(&(ptr->outgoing_commands));
	if(ptr->in_partial!=NULL)
	{
		g_string_free(ptr->in_partial,TRUE);
		ptr->in_partial=NULL;
	}
	if(ptr->created_lock!=NULL)
		free(ptr->created_lock);
	if(ptr->temp_nick!=NULL)
		free(ptr->temp_nick);
	if(ptr->user_pass!=NULL)
		free(ptr->user_pass);
	bzero(ptr, sizeof(CNX_ENTRY));
	ptr->sock_fd=-1;
	route_table->server_count--;
	/* recalcule the max // FIXME */
	printf(COLOYEL1"end lost hub"COLONONE"\n");
	return 0;
}










