/* DChub - a Direct Connect Hub for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * multi_private_chat.c: Copyright (C) Uriel Chemouni <uriel@adess.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "gvar.h"
#include "cnx_lst.h"
#include "multi_public_chat.h"

typedef struct s_cnx_link
{
	int		chat_id;
	int		link_id;
	CNX_ENTRY	*entry;
	struct s_cnx_link *next;
}CNX_LINK;

typedef struct
{
	char		*chat_name;	/* used in the From: field */
	CNX_LINK	**user_lst;
	int		allocated;
	int		len;
	int		chat_id;
} PUBCHAT;

typedef struct s_gl_pubchat
{
	PUBCHAT	*table;
	int		allocated;
	int		len;
}GL_PUBCHAT;


static GL_PUBCHAT *pubchat=NULL;		/* array of PUBCHAT */

static int	get_free_pubchat_id(void)
{
	int	i;

	if(pubchat==NULL)
	{
		pubchat = malloc(sizeof(GL_PUBCHAT));
		pubchat->allocated = 64;
		pubchat->table = malloc(pubchat->allocated * sizeof(PUBCHAT));
		bzero(pubchat->table,pubchat->allocated * sizeof(PUBCHAT));
		return (0);
	}
	for (i=0; i < pubchat->allocated; i++)
	{
		if (pubchat->table[i].chat_name == NULL)
			return i;
	}

	/* realloc */
	i = pubchat->allocated;
	pubchat->allocated *= 2;
	pubchat->table = realloc(pubchat->table, pubchat->allocated * sizeof(PUBCHAT));
	bzero(pubchat->table + i, i * sizeof(PUBCHAT));
	return (i);
}

/******************************************/
/* check if the given name is a chat_name */
/******************************************/
/* output: 1=yes, 0=no */
/***********************/
int is_a_pubchat_name(const char *name)
{
	if(pubchat != NULL)
	{
		int i;

		for(i = 0; i < pubchat->allocated; i++)
		{
			if ((pubchat->table[i].chat_name != NULL) && (!strcmp(name, pubchat->table[i].chat_name)))
				return (1);
		}
	}
	return 0;
}

/*******************************************************/
/* find the index of the given chat in the pubchat array */
/*******************************************************/
int find_pubchat_idx_by_name(const char *chat_name)
{
	int i;

	if(pubchat == NULL)
		return -1;

	for(i = 0; i < pubchat->allocated; i++)
	{
		if((pubchat->table[i].chat_name) &&
			(!strcmp(chat_name,pubchat->table[i].chat_name)))
			return i;
	}
	return -1;
}

/******************************************************/
/* find the index of the given user in the given chat */
/******************************************************/

static int find_pubuser_idx_by_ent(const CNX_ENTRY *ptr, const PUBCHAT *chat)
{
	CNX_LINK	*link;

	if (ptr)
	{
		for (link = ptr->next_on_chat; link; link = link->next)
		{
			if (link->chat_id == chat->chat_id)
				return (link->link_id);
		}
	}
	return -1;
}

static int find_pubuser_idx_by_ent_in(const CNX_ENTRY *ptr, int chat_id)
{
	if (ptr)
	{
		CNX_LINK	*link;

		for (link = ptr->next_on_chat; link; link = link->next)
		{
			if (link->chat_id == chat_id)
				return (link->link_id);
		}
	}
	return -1;
}


/******************************************************/
/* join a chan														*/
/* cmd == "+list"													*/
/* ptr is the user making the command						*/
/* dest_base="$To: remotenick From: Hub-Security $"	*/
/* str=the full incoming string as it as been entered */
/* xtra_param= NULL												*/
/* splitted_str= str splitted along " "					*/

int	cmd_list_multi_pubchat(const char *cmd, CNX_ENTRY *ptr,GString **dest_base, const GString *str, char *xtra_param, const GPtrArray *splitted_str)
{
	int	i;

	printf ("list call\n");
	g_string_sprintf((*dest_base),"<Kernel> ");
	if (pubchat == NULL)
	{ 
		g_string_sprintfa((*dest_base),"there is no public Chan open.\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		return 0;
	}
	
	for (i = 0; i < pubchat->allocated; i++)
	{
		if (pubchat->table[i].chat_name != 0)
			g_string_sprintfa((*dest_base),"\r\n%s with %d users.", pubchat->table[i].chat_name, pubchat->table[i].len);
	}
	g_string_sprintfa((*dest_base),"|");
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}
/******************************************************/
/* join a chan														*/
/* cmd == "+join "												*/
/* ptr is the user making the command						*/
/* dest_base="$To: remotenick From: Hub-Security $"	*/
/* str=the full incoming string as it as been entered */
/* xtra_param= NULL												*/
/* splitted_str= str splitted along " "					*/

int join_pubchat(const char *cmd, CNX_ENTRY *ptr,GString **dest_base, const GString *str, char *xtra_param, const GPtrArray *splitted_str)
{
	int	idx;
	int	i;

	if (splitted_str->len != 2)
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
		send_const_str_to(ptr,(*dest_base)->str);
		return 0;
	}

	{
		char	*name;
		
		name = splitted_str->pdata[1];
		if ((name[0] != '#') || (strlen(name) >= 47))
		{
			g_string_sprintf((*dest_base),"<Kernel> invalid channel name.|");
			send_const_str_to(ptr,(*dest_base)->str);
			return 0;
		}
	}

	if (find_cnx_by_nickname(splitted_str->pdata[1]))
	{
		/* g_string_sprintfa((*dest_base),"a user with the same name already exists.\r\n|"); */
		g_string_sprintf((*dest_base),"<Kernel> a user with the same name already exists.|");
		send_const_str_to(ptr,(*dest_base)->str);
		return 0;
	}
	if(!strcmp("Hub-Security",splitted_str->pdata[1])) /* and perhaps it is also the special nick "Hub-Security" */
	{
		/* g_string_sprintfa((*dest_base),"This name is protected, you cannot use it.\r\n|"); */
		g_string_sprintf((*dest_base),"<Kernel> This name is protected, you cannot use it.|");
		send_const_str_to(ptr,(*dest_base)->str);
		return 0;
	}

	idx=find_pubchat_idx_by_name(splitted_str->pdata[1]);

	/*
	 * The chan do not exist create it.
	 */

	if (idx < 0)
	{
		GString *output;
		int multi_allow;
		int ret;

		G_LOCK(conf_file);
		ret=e_db_int_get(conf_file,"PUBCHAT",&multi_allow);
		G_UNLOCK(conf_file);

		if(ret==0)
		{
			g_string_sprintfa((*dest_base),"PUBCHAT key (int) does not exist, multichat is disabled\r\n|");
			send_const_str_to(ptr,(*dest_base)->str);
			return 0;
		}
		switch(multi_allow)
		{
			case 0:		/* disabled for everyone */
						return 0;
		
			case 1:	/* op and master only */
						if (!(ptr->is_op & OP_PRIV))
						{
							g_string_sprintfa((*dest_base),"no such chan\r\n|");
							send_const_str_to(ptr,(*dest_base)->str);
							return 0;
						}
						break;
		
			case 2:		/* allowed for everyone */
						break;
		}
		idx = get_free_pubchat_id();
		if (idx == -1)
		{
			g_string_sprintfa((*dest_base),"critical error.\r\n|");
			send_const_str_to(ptr,(*dest_base)->str);
			return 0;
		}

		/*	init the entry */
		pubchat->table[idx].chat_name=strdup(splitted_str->pdata[1]);
		pubchat->table[idx].allocated = 32;
		pubchat->table[idx].len = 0;
		pubchat->table[idx].chat_id = idx;
		pubchat->table[idx].user_lst=malloc(pubchat->table[idx].allocated * sizeof(CNX_LINK*));
		bzero(pubchat->table[idx].user_lst, pubchat->table[idx].allocated * sizeof(CNX_LINK*));

		/* and send a message to this user on this newly create chat */
		output=g_string_new("");
		g_string_sprintf(output,"$To: %s From: %s $Multi-users chat created\r\n|",ptr->user_nick->str,
					 pubchat->table[idx].chat_name);
		send_const_str_to(ptr,output->str);
		g_string_free(output,TRUE);
	}
	if (find_pubuser_idx_by_ent_in(ptr, idx) != -1)
	{
		g_string_sprintfa((*dest_base),"You are already in this channel\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		return 0;
	}
	if (pubchat->table[idx].allocated <= pubchat->table[idx].len)
	{
		int	old_size;
		
		old_size = pubchat->table[idx].allocated;
		pubchat->table[idx].allocated *= 2;
		pubchat->table[idx].user_lst = realloc(pubchat->table[idx].user_lst, pubchat->table[idx].allocated * sizeof(CNX_LINK*));
		bzero(pubchat->table[idx].user_lst + old_size, old_size * sizeof(CNX_LINK*));
	}

	for (i = 0; pubchat->table[idx].user_lst[i] != NULL; i++)
		;

	pubchat->table[idx].user_lst[i] = malloc(sizeof(CNX_LINK));
	pubchat->table[idx].user_lst[i]->entry = ptr;
	pubchat->table[idx].user_lst[i]->next = ptr->next_on_chat;
	pubchat->table[idx].user_lst[i]->link_id = i;
	pubchat->table[idx].user_lst[i]->chat_id = idx;
	ptr->next_on_chat = pubchat->table[idx].user_lst[i];
	pubchat->table[idx].len++;
	{
		GString	*output;

		output=g_string_new("");
		g_string_sprintf(output,"$Hello %s|",pubchat->table[idx].chat_name);
		send_const_str_to(ptr,output->str);

		g_string_sprintf(output,"$To: %s From: %s $now you can speak\r\n|",
				 ptr->user_nick->str,
				 pubchat->table[idx].chat_name);
		send_const_str_to(ptr,output->str);

		g_string_sprintf(output,"user %s join the chan\r\n|",ptr->user_nick->str);
		send_to_users_of_multi_pubchat_except2(idx, output->str, ptr);

		g_string_free(output,TRUE);
	}
	return (0);
}

void	append_public_chat_name(GString *ulst, CNX_ENTRY *ptr)
{
	CNX_LINK *lnk;
	
	for (lnk = ptr->next_on_chat; lnk != 0; lnk = lnk->next)
		g_string_sprintfa(ulst,"%s$$", pubchat->table[lnk->chat_id].chat_name);
}

int pubchat_user_quit(CNX_ENTRY *ptr)
{
	CNX_LINK *lnk;
	CNX_LINK *next_lnk;

	for (lnk = ptr->next_on_chat; lnk != 0; lnk = next_lnk)
	{
		next_lnk = lnk->next;
		{
			GString	*output;

			output=g_string_new("");
	
			g_string_sprintf(output,"Timeout user %s lost connection to the chan\r\n|",ptr->user_nick->str);
			send_to_users_of_multi_pubchat_except2(lnk->chat_id, output->str, ptr);
			g_string_free(output,TRUE);
		}
		pubchat->table[lnk->chat_id].user_lst[lnk->link_id] = NULL; /* should be equal to lnk before affectation */
		pubchat->table[lnk->chat_id].len--;
			
		if (pubchat->table[lnk->chat_id].len == 0) /* destroy chan if empty */
		{
			int	tmp;
		
			tmp = lnk->chat_id;
			pubchat->len--;
			free(pubchat->table[tmp].chat_name);
			pubchat->table[tmp].chat_name = NULL;
			free (pubchat->table[tmp].user_lst);
			pubchat->table[tmp].user_lst = NULL;
		}
		free(lnk);
	}
	return 0;
}

#if 0
int pubchat_user_leave(CNX_ENTRY *ptr, int chat_id)
{
	CNX_LINK *lnk;
	int	uidx;

	uidx = find_pubuser_idx_by_ent_in(ptr, chat_id);
	lnk = pubchat->table[chat_id].user_lst[uidx];
	pubchat->table[chat_id].user_lst[uidx] = NULL;
	free(lnk);
	pubchat->table[lnk->chat_id].len--;
	if (pubchat->table[lnk->chat_id].len == 0) /* destroy chan if empty */
	{
		int	tmp;
			
		tmp = lnk->chat_id;
		pubchat->len--;
		
		free (pubchat->table[tmp].chat_name);
		pubchat->table[tmp].chat_name = NULL;
		free (pubchat->table[tmp].user_lst);
		pubchat->table[tmp].user_lst = NULL;
	}
	else
	{
		GString	*output;

		g_string_sprintf(output,"user %s leave the chan\r\n|",ptr->user_nick->str);
		send_to_users_of_multi_pubchat_except2(idx, output->str, ptr);
		g_string_free(output,TRUE);
	}
	return (0);
}
#endif

/****************************************************************************************/
/* create "$To: destnick From: chatnick $base_msg" and send it to all users of the chat */
/****************************************************************************************/
static void send_to_users_of_multi_pubchat_except(PUBCHAT *m,char *base_msg, int skip_idx)
{
	GString *tmp;
	GString_ref *tmp2;
	int i;

	tmp = g_string_sized_new(128);
	g_string_sprintf(tmp, "$To: %s From: %s $%s|", m->chat_name, m->chat_name, base_msg);
	tmp2 = conv_gstring_to_gstring_ref(tmp);
	for(i=0;i < m->allocated; i++)
	{
		if (i == skip_idx)		/* skip this user */
			continue;

		if(m->user_lst[i] != NULL)
		{
			send_str_to(m->user_lst[i]->entry,tmp2);
		}
	}
	g_string_ref_free(tmp2);
}

void	send_to_users_of_multi_pubchat_except2(int idx, char *base_msg, CNX_ENTRY *ptr)
{
	PUBCHAT *m = pubchat->table + idx;
	int	skip_idx;

	if (ptr == NULL)
	{
		skip_idx = -1;
	}
	else
	{
		skip_idx = find_pubuser_idx_by_ent(ptr, m);
		if (skip_idx == -1)
		{
			/* the user is not allow to speak here */
			return;
		}
	}
	send_to_users_of_multi_pubchat_except(m, base_msg, skip_idx);
}

/***********************************/
/* leave the given multi user chat */
/**************************************/
/* cmd syntax: "+leave_pubchat chatname */
/******************************************************/
/* cmd == "+leave_pubchat "									*/
/* ptr is the user making the command						*/
/* dest_base="$To: remotenick From: Hub-Security $"	*/
/* str=the full incoming string as it as been entered	*/
/* xtra_param= NULL												*/
/* splitted_str= str splitted along " "					*/
/******************************************************************************/
/* the function must return >=0 if the connection must remain alive (no error */
/* (0) or recoverable error (>0)) and a value <0 if there is a fatal error)	*/
/******************************************************************************/
int cmd_leave_multi_pubchat(const char *cmd, CNX_ENTRY *ptr,GString **dest_base, const GString *str, char *xtra_param, const GPtrArray *splitted_str)
{
	int ret=0;
	int idx;
	int uidx;
	PUBCHAT *chat;

	if(splitted_str->len!=2)	/* cmd + 1 param */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s. Must be 1\r\n|",cmd);
		send_const_str_to(ptr,(*dest_base)->str);
		return ret;
	}

	idx=find_pubchat_idx_by_name(splitted_str->pdata[1]);
	if(idx==-1)
	{
		g_string_sprintfa((*dest_base),"no chat with this name exists\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	chat=pubchat->table + idx;
	uidx=find_pubuser_idx_by_ent(ptr,chat);
	if(uidx==-1)
	{
		g_string_sprintfa((*dest_base),"you are not member of this chat\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	{	 /* remove the user from this chat */
		CNX_LINK	*tmp;

		if (ptr->next_on_chat == chat->user_lst[uidx])
		{
			ptr->next_on_chat = chat->user_lst[uidx]->next;
			free(chat->user_lst[uidx]);
			chat->user_lst[uidx] = NULL;
		}
		else
		{
			for (tmp = ptr->next_on_chat; tmp->next; tmp=tmp->next)
			{
				if (tmp->next->chat_id == idx)
				{
					CNX_LINK	*tmp2;

					tmp2 = tmp->next;
					pubchat->table[tmp2->chat_id].user_lst[tmp2->link_id] = 0;
					tmp->next = tmp2->next;
					free(tmp2);
					break;
				}
			}
		}
	}

	/* and notify this change to other user */
	{
		gchar *tmp;

		tmp=g_strconcat(ptr->user_nick->str," has left the chat.\r\n|",NULL);
		send_to_users_of_multi_pubchat_except(chat,tmp,-1);
		g_free(tmp);
	}

	{ /* force the "fake" user to be undisplayed */
		GString *output;

		output=g_string_new("");
		g_string_sprintf(output,"$Quit %s|",chat->chat_name);
		send_const_str_to(ptr,output->str);
		
		g_string_sprintf(output,"$To: %s From: %s $You have left this chat.\r\n|",ptr->user_nick->str,chat->chat_name);
		send_const_str_to(ptr,output->str);

		g_string_sprintf(output,"user %s leave the chan\r\n|",ptr->user_nick->str);
		send_to_users_of_multi_pubchat_except2(idx, output->str, NULL);

		g_string_free(output,TRUE);
	}
	chat->len--;
	/* and delete the chat if it becomes empty */
	if(chat->len == 0)
	{
		/* there is no one on this chat, destroy it */
		free(chat->chat_name);
		chat->chat_name = NULL;
		pubchat->len--;
	}
	
 abrt:
	return ret;
}

